/*
 * Copyright (c) 2004-2007 Erez Zadok
 * Copyright (c) 2004-2007 Stony Brook University
 * Copyright (c) 2004-2007 The Research Foundation of State University of New York
 * Copyright (c) 2004-2007 Nikolai Joukov
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package, or get one from ftp://ftp.filesystems.org/pub/fist/COPYING.
 *
 * This Copyright notice must be kept intact and distributed with all
 * fistgen sources INCLUDING sources generated by fistgen.
 */

#include "replay.h"
#include "replayk.h"
#include "replay_timing.h"
#include <linux/delay.h>
#include <linux/time.h>

/*
 * define MJITTER in the Makefile to measure the timing jitter
 * (for debugging)
 */

/* enable pre-spin */
#ifdef REPLAY_TIMING_PRESPIN
#define PRESPIN
#endif

static struct timeval real_start_time; //Actual start time
static timeval_t start_time; //Start Timestamp of the original op
static timeval_t total_time;
static timeval_t prev_time; //The timestamp of the last op
/*static timeval_t gross_err_time;
static s_timeval_t net_err_time;
static timeval_t last_err_time; //The error on the last op
static timeval_t total_sleep_time; //Total sleep time */
static struct timeval pre_op_time;

/*static int replay_timing_add_error(long secs_err,long usecs_err)
{
  unsigned long secs,usecs;
  print_entry_location();

  if(secs_err >= 0)
    secs = secs_err;
  else
    secs = secs_err * -1;

  if(usecs_err >= 0)
    usecs = usecs_err;
  else
    usecs = usecs_err * -1;

  last_err_time.secs = secs;
  last_err_time.usecs = usecs;

  //Gross Error
  gross_err_time.secs += secs;
  gross_err_time.usecs += usecs;
  if(gross_err_time.usecs + usecs >= 1000000) {
    gross_err_time.secs++;
    gross_err_time.usecs = (gross_err_time.usecs + usecs) - 1000000;
  }

  //Net Error
  net_err_time.secs += secs;
  net_err_time.usecs += usecs_err;
  if(net_err_time.usecs >= 1000000) {
    net_err_time.secs++;
    net_err_time.usecs -= 1000000;
  }
  else if(net_err_time.usecs < 0) {
    net_err_time.secs--;
    net_err_time.usecs += 1000000;
  }

  print_exit_location();
  return 0;
}*/

/*
 * Timing: sleep to emulate timing correctly in between OPS.
 */
static inline void replay_timing_accurately_in_between_ops(timeval_t* time_prev,timeval_t* time_current,int speed)
{
  int sleep = 0;
  long elapsed;
  long total;
  struct timeval now;
  print_entry_location();

  do_gettimeofday(&now);

  //Convert to milliseconds
  elapsed = (now.tv_sec - pre_op_time.tv_sec) * 1000;
  elapsed += now.tv_usec / 1000 - pre_op_time.tv_usec / 1000;

  total = elapsed;

  sleep = (time_current->secs - time_prev->secs) * 1000;
  sleep += time_current->usecs / 1000 - time_prev->usecs / 1000;
  sleep -= elapsed;

  //Adjust Speed - Speed is a percent of the original
  if(speed != 100) {
    if(sleep > 0) {
      if(speed < 100) {
	if(sleep * (100 - speed) < sleep) //overflow
	  sleep += (sleep / 100) * (100 - speed);
	else
	  sleep += (sleep * (100 - speed)) / 100;
      }
      else {
	if(sleep * 100 < speed) //overflow
	  sleep = (sleep / speed) * 100;
	else
	  sleep = (sleep * 100) / speed;
      }
    }
  }

  if(sleep > 0) {
    total += sleep;
    msleep(sleep);
  }

  do_gettimeofday(&pre_op_time);
  print_exit_location();
}

static inline void replay_timing_accurately_elapsed(timeval_t* time_op, int speed)
{
  long sleep = 0;
  long elapsed;
  long real_elapsed;
  long total;
  struct timeval now;

  print_entry_location();

  do_gettimeofday(&now);

  //Convert to milliseconds
  real_elapsed = (now.tv_sec - real_start_time.tv_sec) * 1000;
  real_elapsed += (now.tv_usec - real_start_time.tv_usec) / 1000;
  elapsed = (time_op->secs - start_time.secs) * 1000;
  elapsed += (time_op->usecs / 1000 - start_time.usecs / 1000);

  //Adjust Speed - Speed is a percent of the original
  if(speed != 100) {
    if(speed < 100) {
      if(elapsed * (100 - speed) < elapsed) //overflow
	elapsed += (elapsed / 100) * (100 - speed);
      else
	elapsed += (elapsed * (100 - speed)) / 100;
    }
    else {
      if(elapsed * 100 < elapsed) //overflow
	elapsed = (elapsed / speed) * 100;
      else
	elapsed = (elapsed * 100) / speed;
    }
  }

  sleep = elapsed - real_elapsed;
  total = real_elapsed;
  if(sleep > 0) {
    total += sleep;
    msleep(sleep);
  }

  print_exit_location();
}

#ifdef MJITTER
extern void mjitter_add_value(unsigned int error);
#endif
/*
 * Here we assume that the traces that we replay are shorter than 40G usecs
 * (40G usecs ~= 5 hours )
 * As a benefit we use 32 bit values in math.
 */
static inline void replay_timing_accurately_elapsed_prespin(timeval_t* time_op, int speed)
{
	long sleep = 0;
	int trace_elapsed;
	int now_elapsed;
	struct timeval now;

	print_entry_location();

	sleep = time_op->secs  - start_time.secs;
	trace_elapsed =  (int)sleep * 100000; // This is why we are limited by 5 hours...
	trace_elapsed += ((int)time_op->usecs - (int)start_time.usecs) / 10;
	do_gettimeofday(&now);
	sleep = now.tv_sec  - real_start_time.tv_sec;
	now_elapsed =  (int)sleep * 100000;
	now_elapsed += ((int)now.tv_usec - (int)real_start_time.tv_usec) / 10;

	if (trace_elapsed > now_elapsed) {
		if (speed == 100)
			sleep = (trace_elapsed - now_elapsed) / 100;
		else {
			sleep = trace_elapsed - now_elapsed;
			sleep = (sleep * speed) / 10000;
		}
#ifdef PRESPIN
		if (sleep > 1)
			msleep(sleep);
#else
		if (sleep > 0)
			msleep(sleep);
#endif
	}

#if defined(PRESPIN) || defined(MJITTER)
	if (speed != 100) {
		sleep = (trace_elapsed * speed) / 100;
		trace_elapsed = (int)sleep;
	}
#endif

#ifdef PRESPIN
	// busy wait (pre-spin)
	do {
#endif

#if defined(PRESPIN) || defined(MJITTER)
		do_gettimeofday(&now);
		sleep = now.tv_sec  - real_start_time.tv_sec;
		now_elapsed =  (int)sleep * 100000;
		now_elapsed += ((int)now.tv_usec - (int)real_start_time.tv_usec) / 10;
		if (speed != 100) {
			sleep = (now_elapsed * speed) / 100;
			now_elapsed = (int)sleep;
		}
#endif

#ifdef PRESPIN
	/* allow 10 microseconds for the function return delay etc. */
	} while (now_elapsed < trace_elapsed - 1);
#endif

#ifdef MJITTER
	if (trace_elapsed > now_elapsed)
		mjitter_add_value(trace_elapsed - now_elapsed);
	else
		mjitter_add_value(now_elapsed - trace_elapsed);
#endif

	print_exit_location();
}

/*
 * Timing: put kernel thread to sleep in between Ops to accurately
 * emulate traces.
 */
int replay_timing_pre_op(struct replay_command* rcmdp,int mode,int speed)
{
	int err = 0;

	print_entry_location();

	if(!prev_time.secs && !prev_time.usecs) {
		//First op
		start_time = rcmdp->time;
		goto out;
	}

	switch(mode) {
		case REPLAY_TIMING_OFF:
			break;
		case REPLAY_TIMING_BETWEEN_OPS:
			replay_timing_accurately_in_between_ops(&prev_time,&rcmdp->time, speed);
			break;
		case REPLAY_TIMING_ELAPSED:
		  replay_timing_accurately_elapsed_prespin(&rcmdp->time, speed);
			break;
		default:
			print_error_message("Unsupported Timing Mode");
			err = -1;
			break;
	}

out:
	prev_time = rcmdp->time;
	print_exit_location();
	return err;
}

/*int replay_timing_post_op()
{
  print_entry_location();

  print_exit_location();
  return 0;
}*/

//extern int mjitter_init(void);
int replay_timing_init()
{
	int err = 0;
	print_entry_location();

	do_gettimeofday(&real_start_time);

	memset(&prev_time,0,sizeof(timeval_t));
	memset(&start_time,0,sizeof(timeval_t));
//	memset(&total_sleep_time,0,sizeof(timeval_t));
	memset(&pre_op_time,0,sizeof(struct timeval));

	/*#ifdef MJITTER
	err = mjitter_init();
	#endif*/

	print_exit_status(err);
	return err;
}

//extern void mjitter_release(void);
void replay_timing_release()
{
  /*#ifdef MJITTER
	print_entry_location();
	mjitter_release();
	print_exit_location();
	#endif*/
}

/*timeval_t replay_timing_get_gross_error()
{
  if(REPLAY_TIMING_ELAPSED) {
    timeval_t time;
    time.usecs = 0;
    time.secs = 0;
    return time;
  }
  return gross_err_time;
}

s_timeval_t replay_timing_get_net_error()
{
  if(REPLAY_TIMING_ELAPSED) {
    s_timeval_t time;
    time.usecs = 0;
    time.secs = 0;
    return time;
  }
  return net_err_time;
}

timeval_t replay_timing_get_total_time()
{
  return total_time;
}

timeval_t replay_timing_get_elapsed_time()
{
  struct timeval now;
  timeval_t elapsed_time;
  long usecs;

  do_gettimeofday(&now);
  elapsed_time.secs = now.tv_sec - real_start_time.tv_sec;
  usecs = now.tv_usec - real_start_time.tv_usec;

  if(usecs < 0) {
    elapsed_time.secs--;
    usecs += 1000000;
  }
  elapsed_time.usecs = usecs;

  return elapsed_time;
}

inline timeval_t replay_timing_get_total_sleep_time()
{
  return total_sleep_time;
}*/

