/*
 * Copyright (c) 2003-2007 Erez Zadok
 * Copyright (c) 2003-2007 Stony Brook University
 * Copyright (c) 2003-2007 The Research Foundation of State University of New York
 * Copyright (c) 2004-2007 Nikolai Joukov
 * Copyright (c) 2003-2005 Akshat Aranya
 * Copyright (c) 2006-2007 Gopala Surya
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package, or get one from ftp://ftp.filesystems.org/pub/fist/COPYING.
 *
 * This Copyright notice must be kept intact and distributed with all
 * fistgen sources INCLUDING sources generated by fistgen.
 */
/*
 * This file defines structures and constants for expression graph
 * generator.
 */

#ifndef __GENGRAPH_H
#define __GENGRAPH_H


#define PCAP_ERRBUF_SIZE 256


struct edge {
	struct block *succ;
	struct block *pred;
};

struct block {
    int type;
    union {
	int ivalue;
	char *svalue;
    } u;
    int sense;
    struct block *head;
    struct edge et;
    struct edge ef;
    int index;
};

struct graphset {
	struct block *root;
	int n_ods;
	struct user_output_driver *ods;
	int n_ads;
	struct user_assembly_driver *ads;
	struct graphset *next;
};

struct array {
	int count;
	void *data;
};

#define JF(b) ((b)->ef.succ)
#define JT(b) ((b)->et.succ)

struct block *gen_int_block(int, int);
struct block *gen_string_block(int, char *);
struct block *gen_retblk(int);

char *gen_bmap(int);
char *set_bit(char *, int);
struct block *gen_bmap_block(int, char *);

void gen_and(struct block *, struct block *);
void gen_or(struct block *, struct block *);
void gen_not(struct block *);

char *sdup(const char *);

struct block *finish_parse(struct block *);

struct user_output_driver *gen_file_destination(char *, int, int sync_writes);
struct user_output_driver *gen_socket_destination(char *hostname, int port, int buf);
struct user_output_driver *gen_async_destination(char *, int, int, int sync_writes);
struct user_output_driver *gen_buffer_destination(int);
struct user_output_driver *gen_encrypt_destination(char *, int ishexstr);
struct user_output_driver *gen_cksum_destination(char *key, int ishexstr);
struct user_output_driver *gen_compress_destination();
struct user_assembly_driver *gen_assembly(int type, int flags);

int add_destination(struct graphset **graphset, struct block *root, struct array *adlist, struct array *odlist);

struct array *append_array(struct array *ar, void *item, int elsize);
struct array *make_array(void *item, int elsize);

#ifdef _DEBUG
#define debug(format, ...) printf(format, ## __VA_ARGS__)
#else
#define debug(format, args...)
#endif

struct graphset *tracefs_compile();
void tracefs_compile_done();
#endif
